<?php
/**
 * Author: lopiu
 * Date: 03.07.2020
 * Time: 07:24
 */

namespace Battles;

use Exceptions\GameException;
use Battles\Database\Db;
use Throwable;
use Battles\Models\BankModel;

class Bank
{
    public int $user_id = 0;
    private int $money = 0;
    private $user;

    private const ERROR_NO_MONEY_IN_WALLET = "Ошибка! Нет денег в кошельке!";
    private const ERROR_NO_BANK_ACCOUNT = "Ошибка! Счёта не существует!";
    private const ERROR_NO_MONEY_IN_BANK_ACCOUNT = "Ошибка! Нет денег на счету!";
    private const ERROR_WRONG_AMOUNT = "Ошибка! Сумма должна быть положительной!";
    private const MINIMUM_COMISSION_VALUE = 1;
    private const OPERATION_SEND_MONEY = 'sendMoney';
    private const OPERATION_DEPOSIT_MONEY = 'depositMoney';
    private const OPERATION_RECEIVE_MONEY = 'receiveMoney';
    private const OPERATION_WITHDRAW_MONEY = 'withdrawMoney';
    private const LOG = [
        self::OPERATION_SEND_MONEY => 'Банк: Перевод средств на другой счёт.',
        self::OPERATION_RECEIVE_MONEY => 'Банк: Получение средств.',
        self::OPERATION_DEPOSIT_MONEY => 'Пополнение счёта.',
        self::OPERATION_WITHDRAW_MONEY => 'Снятие денег со счёта.',
        'clanRegister' => 'Оплата стоимости регистрации клана.',
        'sellShop' => 'Продажа товара в магазине.'
    ];

    public function __construct(int $user_id)
    {
        $bank_row = Db::getInstance()->fetch('SELECT user_id, money FROM bank WHERE user_id = ?', $user_id);
        $this->user = Db::getInstance()->fetch('SELECT money FROM users WHERE id = ?', $user_id);
        foreach ($this as $key => $value) {
            if (isset($bank_row[$key])) {
                $this->$key = $bank_row[$key];
            }
        }
    }

    /**
     * Комиссия: процент от переводимой суммы, но не менее 1 кр. Задаётся в config.php.
     *
     * @param int $amount сумма.
     *
     * @return int
     */
    private function bankCommission(int $amount): int
    {
        $bankCommission = round($amount * GameConfigs::BANK_COMISSION);
        return $bankCommission >= self::MINIMUM_COMISSION_VALUE ? (int)$bankCommission : self::MINIMUM_COMISSION_VALUE;
    }

    /**
     * Пишем банковское событие в лог в БД
     *
     * @param int $receiverId ID получателя.
     * @param int $amount сумма.
     * @param string $operationType тип банковской операции.
     * @param ?int $senderId ID отправителя (ID игрока, если не указано иное).
     *
     * @return void
     */
    private function bankLogs(int $receiverId, int $amount, string $operationType, ?int $senderId = null): void
    {
        if (is_null($senderId)) {
            $senderId = $this->user_id;
        }
        $text = self::LOG[$operationType];
        if ($operationType == self::OPERATION_SEND_MONEY) {
            $text .= " Комиссия: " . $this->bankCommission($amount);
        } elseif ($operationType == self::OPERATION_DEPOSIT_MONEY) {
            $receiverId = $this->user_id;
        } elseif ($operationType == self::OPERATION_WITHDRAW_MONEY) {
            $receiverId = $this->user_id;
            $text .= " Комиссия: " . $this->bankCommission($amount);
        }
        GameLogs::addBankLog($senderId, $receiverId, $amount, $operationType, $text);
    }

    /**
     * Перевод денег между банковскими счетами игроков с банковской комиссией.
     *
     * @param int $receiver ID получателя.
     * @param int $amount сумма.
     *
     * @return int
     * @throws GameException
     */
    public function sendMoney(int $receiver, int $amount): int
    {
        $receiverWallet = Db::getInstance()->ofetch('SELECT money FROM bank WHERE user_id = ?', $receiver);
        if ($amount <= 0) {
            throw new GameException(self::ERROR_WRONG_AMOUNT);
        }
        if (!$receiverWallet) {
            throw new GameException(self::ERROR_NO_BANK_ACCOUNT);
        }
        $amountWithComission = $amount + $this->bankCommission($amount);
        if ($amountWithComission > $this->money) {
            throw new GameException(self::ERROR_NO_MONEY_IN_BANK_ACCOUNT);
        }
        // Снимаем сумму с комиссией у отправителя
        $this->money -= $amountWithComission;
        self::setBankMoney($this->money, $this->user_id);
        $this->bankLogs($receiver, $this->money, self::OPERATION_SEND_MONEY);
        // Отдаём сумму на счёт получателю
        $receiverWallet->money += $amount;
        self::setBankMoney($receiverWallet->money, $receiver);
        $this->bankLogs($receiver, $receiverWallet->money, self::OPERATION_RECEIVE_MONEY);
        // Возвращаем изменившиеся значения
        return $this->money;
    }

    /**
     * Пополнение банковского счёта игрока
     *
     * @param int $amount сумма.
     *
     * @return array
     * @throws GameException
     */
    public function depositMoney(int $amount): array
    {
        if ($amount <= 0) {
            throw new GameException(self::ERROR_WRONG_AMOUNT);
        }
        $walletMoney = Db::getInstance()->fetchColumn('SELECT money FROM users WHERE id = ?', $this->user_id);
        if ($walletMoney < $amount) {
            throw new GameException(self::ERROR_NO_MONEY_IN_WALLET);
        }
        // Забираем деньги из кошелька получателя
        $this->user->money -= $amount;
        self::setWalletMoney($this->user->money, $this->user_id);
        // Отдаём сумму на счёт получателю
        $this->money += $amount;
        self::setBankMoney($this->money, $this->user_id);
        $this->bankLogs(0, $this->money, self::OPERATION_DEPOSIT_MONEY);
        // Возвращаем изменившиеся значения
        return [
            'walletMoney' => $this->user->money,
            'bankMoney' => $this->money
        ];
    }

    /**
     * Снятие денег с банковского счёта игрока с банковской комиссией.
     *
     * @param int $amount сумма.
     *
     * @return array
     * @throws GameException
     */
    public function withdrawMoney(int $amount): array
    {
        if ($amount <= 0) {
            throw new GameException(self::ERROR_WRONG_AMOUNT);
        }
        $amountWithComission = $amount + $this->bankCommission($amount);
        if ($this->money < $amountWithComission) {
            throw new GameException(self::ERROR_NO_MONEY_IN_BANK_ACCOUNT);
        }
        // Снимаем сумму с комиссией у отправителя
        $this->money -= $amountWithComission;
        self::setBankMoney($this->money, $this->user_id);
        $this->bankLogs(0, $this->money, self::OPERATION_WITHDRAW_MONEY);
        // Отдаём сумму в кошелёк получателя
        $this->user['money'] += $amount;
        self::setWalletMoney($this->user['money'], $this->user_id);
        // Возвращаем изменившиеся значения
        return [
            'walletMoney' => $this->user['money'],
            'bankMoney' => $this->money
        ];
    }

    /**
     * Установить количество денег на банковском счету.
     *
     * @param int $amount сумма.
     * @param int $uid ID пользователя.
     * @param string $operationType Тип операции. По умолчанию пусто. Если ввести, система запишет событие в банковский лог.
     *
     * @return void
     */
    public static function setBankMoney(int $amount, int $uid, string $operationType = ''): void
    {
        $bm = new BankModel();
        try {
            $bm->setMoney($amount, $uid);
            if ($operationType) {
                GameLogs::addBankLog(0, 0, $amount, $operationType, self::LOG[$operationType]);
            }
        } catch (Throwable $e) {
            echo "Не отработал запрос в БД в файле {$e->getFile()}({$e->getLine()})";
        }

    }

    /**
     * Установить количество денег на руках.
     *
     * @param int $amount сумма.
     * @param int $uid ID пользователя.
     *
     * @return void
     */
    public static function setWalletMoney(int $amount, int $uid): void
    {
        User::getInstance($uid)->setMoney($amount);
        User::getInstance($uid)->saveMoney();
    }

    public function getMoney(): int
    {
        return $this->money;
    }

    public function setMoney($amount)
    {
        $this->money = $amount;
    }
}